//     =========================== OpticalFlow.h ============================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 09.05.2009
//     **********************************************************************

#if !defined(_OpticalFlow_H_INCLUDED_)
#define _OpticalFlow_H_INCLUDED_

#define cOF_ApertureSize 80
#define cOF_Iterations   10
#define cOF_VMapTotal    3
#define cOF_SearchRadius (cOF_ApertureSize/4)
#define cOF_SearchStep   (cOF_ApertureSize/cOF_Iterations)
#define cOF_iCellTotal   (cOF_ApertureSize*4)
#define cOF_iStep        (float(cOF_SearchRadius*2)/float(cOF_iCellTotal))

#define cPI 3.1415926535897932384626433832795

// Class of "Optical flow" algorithm
class CvOpticalFlow {

	// 24-bit colour
	struct CvBGR { uchar b,g,r; };
	
	CvSize mImgSize;
	int    mW,mH,mVMapIdx;
	CvMat* mpDxVMap[cOF_VMapTotal];
	CvMat* mpDyVMap[cOF_VMapTotal];
	CvBGR  mColor[cOF_iCellTotal][cOF_iCellTotal];

CvAssociativeMemory32S* mpAM;

public:

	// Class constructor
	CvOpticalFlow(CvSize aImgSize) {

		mpAM = new CvAssociativeMemory32S;

		int i,j;
		float x,y = -cOF_SearchRadius;

		for(i=0; i<cOF_iCellTotal; i++, y += cOF_iStep) {
			x = -cOF_SearchRadius;
			for(j=0; j<cOF_iCellTotal; j++, x += cOF_iStep) {
					float dDist = x*x + y*y;
					float cH = 180.0f*float(1 + atan2(y, x)/cPI);
					float cL = float(1-exp(-dDist*0.04f));
					CvScalar tColor = HLStoRGB(cH, cL, 1.0f);
					mColor[j][i].r = uchar(tColor.val[2]);
					mColor[j][i].g = uchar(tColor.val[1]);
					mColor[j][i].b = uchar(tColor.val[0]);
			}
		}

		mVMapIdx = 0;
		mImgSize = aImgSize;
		mW = aImgSize.width/cOF_ApertureSize;
		mH = aImgSize.height/cOF_ApertureSize;

		int TreeTotal = mW*mH*cOF_Iterations;
		mpAM->Create(cvSize(cOF_ApertureSize, cOF_ApertureSize), 0, TreeTotal, false);

		for(i=0; i<cOF_VMapTotal; i++) {
			mpDxVMap[i] = cvCreateMat(aImgSize.height, aImgSize.width,  CV_32FC1);
			mpDyVMap[i] = cvCreateMat(aImgSize.height, aImgSize.width,  CV_32FC1);
			cvZero(mpDxVMap[i]);
			cvZero(mpDyVMap[i]);
		}
	}

	// Class destructor
	~CvOpticalFlow() {
		if(mpAM) { delete mpAM; }

		for(int i=0; i<cOF_VMapTotal; i++) {
			cvReleaseMat(&mpDxVMap[i]);
			cvReleaseMat(&mpDyVMap[i]);
		}
	}

	// Calculation of velocity
	void Calculate(IplImage* apImg, CvMat** appVx = NULL, CvMat** appVy = NULL) {

		mVMapIdx = ((++mVMapIdx) % (cOF_VMapTotal-1)) + 1;

		cvZero(mpDxVMap[mVMapIdx]);
		cvZero(mpDyVMap[mVMapIdx]);

		mpAM->SetImage(apImg);

		int i,TreeIdx,xi,yi;
		
		for(TreeIdx=0, i=0; i<cOF_Iterations; i++) {
			for(yi=0; yi<mH; yi++) {
				for(xi=0; xi<mW; xi++, TreeIdx++) {

					int x = (i % 2)?(xi*cOF_ApertureSize + i*cOF_SearchStep):
							(mImgSize.width - xi*cOF_ApertureSize - i*cOF_SearchStep);
							
					int y = yi*cOF_ApertureSize + i*cOF_SearchStep;

					if(x + cOF_ApertureSize >= mImgSize.width ||
					   y + cOF_ApertureSize >= mImgSize.height) { continue; }

					mpAM->SetActiveTree(TreeIdx);
					
					int x0,y0;
					CvRect iAr;
					float  dX = 0, dY = 0;

					if(mpAM->GetTotalABases() > 1) {
						long* pData;
						CvRect rFound;

						iAr = cvRect(x, y, cOF_ApertureSize, cOF_ApertureSize);
						bool Stop = mpAM->Read(iAr, &rFound, &pData);

						int x1 = x - cOF_SearchRadius;
						int x2 = x + cOF_SearchRadius;
						int y1 = y - cOF_SearchRadius;
						int y2 = y + cOF_SearchRadius;

						for(y0=y1; y0<y2 && !Stop; y0 += cOF_SearchStep) {
							for(x0=x1; x0<x2 && !Stop; x0 += cOF_SearchStep) {
								iAr = cvRect(x0, y0, cOF_ApertureSize, cOF_ApertureSize);

								if(x0 < 0) { iAr.x = 0; }
								if(y0 < 0) { iAr.y = 0; }

								if(x0 + cOF_ApertureSize >= mImgSize.width)  {
									iAr.x = mImgSize.width  - cOF_ApertureSize;
								}

								if(y0 + cOF_ApertureSize >= mImgSize.height) {
									iAr.y = mImgSize.height - cOF_ApertureSize;
								}

								if(mpAM->Read(iAr, &rFound, &pData)) {
									Stop = true;
								}
							}
						}

						if(Stop) {
							int cX0 = x + cOF_ApertureSize/2;
							int cY0 = y + cOF_ApertureSize/2;
							int cXf = rFound.x + (rFound.width>>1);
							int cYf = rFound.y + (rFound.height>>1);
							dX = float(cXf - cX0); dY = float(cYf - cY0);
							for(y0=0; y0<cOF_ApertureSize; y0++) {
								int Idx = (y+y0)*mImgSize.width + x;
								for(x0=0; x0<cOF_ApertureSize; x0++, Idx++) {
									mpDxVMap[mVMapIdx]->data.fl[Idx] += dX;
									mpDyVMap[mVMapIdx]->data.fl[Idx] += dY;
								}
							}
						}
					}
					
					mpAM->ClearTreeData();
					iAr = cvRect(x, y, cOF_ApertureSize, cOF_ApertureSize);
					long Data = TreeIdx;
					mpAM->Write(iAr, &Data, false);
				}
			}
		}
		
		cvAdd(mpDxVMap[2], mpDxVMap[1], mpDxVMap[0]);
		cvAdd(mpDyVMap[2], mpDyVMap[1], mpDyVMap[0]);
		cvSmooth(mpDxVMap[0], mpDxVMap[0], CV_GAUSSIAN, 7, 7, 7);
		cvSmooth(mpDyVMap[0], mpDyVMap[0], CV_GAUSSIAN, 7, 7, 7);
		cvConvertScale(mpDxVMap[0], mpDxVMap[0], 1.0/cOF_Iterations);
		cvConvertScale(mpDyVMap[0], mpDyVMap[0], 1.0/cOF_Iterations);

		if(appVx) { *appVx = mpDxVMap[0]; }
		if(appVy) { *appVy = mpDyVMap[0]; }
	}

	// Color conversion
	CvScalar HLStoRGB(float h, float l, float s) {
		CvScalar Res;
		float b, g, r;

		if( s == 0 )
			b = g = r = l;
		else
		{
			static const int sector_data[][3]=
				{{1,3,0}, {1,0,2}, {3,0,1}, {0,2,1}, {0,1,3}, {2,1,0}};
			float tab[4];
			int sector;
    
			float p2 = l <= 0.5f ? l*(1 + s) : l + s - l*s;
			float p1 = 2*l - p2;

			h *= 0.016666666666666666f; // h /= 60;
			if( h < 0 )
				do h += 6; while( h < 0 );
			else if( h >= 6 )
				do h -= 6; while( h >= 6 );

			assert( 0 <= h && h < 6 );
			sector = cvFloor(h);
			h -= sector;

			tab[0] = p2;
			tab[1] = p1;
			tab[2] = p1 + (p2 - p1)*(1-h);
			tab[3] = p1 + (p2 - p1)*h;

			b = tab[sector_data[sector][0]];
			g = tab[sector_data[sector][1]];
			r = tab[sector_data[sector][2]];
		}
		
		Res.val[0] = uchar(b*255.0f);
		Res.val[1] = uchar(g*255.0f);
		Res.val[2] = uchar(r*255.0f);
		Res.val[3] = 0;

		return Res;
	}


	// Visualization of motion
	void DrawFlow(CvArr* apImg, bool aColoredMotion) {

		CvMat stub, *img;
		CvMat fstub, *pVx, *pVy;
		int i,j;

		img = cvGetMat( apImg, &stub );
		pVx = cvGetMat( mpDxVMap[0], &fstub );
		pVy = cvGetMat( mpDyVMap[0], &fstub );

		if(aColoredMotion) {
			CvBGR* OutImg_BGR;
			OutImg_BGR = (CvBGR*)img->data.ptr;
			for(long y=0; y<img->height; y++) {
				for(long x=0; x<img->width; x++) {
					long Index = y*img->width + x;
					float  dX =  pVx->data.fl[Index];
					float  dY =  pVy->data.fl[Index];

					int ix = int((cOF_SearchRadius + dX)/cOF_iStep);
					int iy = int((cOF_SearchRadius + dY)/cOF_iStep);

					if(ix < 0) { ix=0; }
					if(iy < 0) { iy=0; }
					if(ix >= cOF_iCellTotal) { ix = cOF_iCellTotal-1; }
					if(iy >= cOF_iCellTotal) { iy = cOF_iCellTotal-1; }
					
					OutImg_BGR[Index] = mColor[ix][iy];
				}
			}
		} else {
			int delta = 10;

			for(i=delta; i < img->rows; i += delta) {
				for(j=delta; j < img->cols; j += delta) {
					CvPoint pt1 = {j, i}, pt2;
					int Idx = pVx->cols*i + j;
					pt2.x = pt1.x + cvRound(pVx->data.fl[Idx]);
					pt2.y = pt1.y + cvRound(pVy->data.fl[Idx]);
					cvLine( img, pt1, pt2, CV_RGB(255,255,255), 1, 8 );
				}
			}
		}
	}
};

#endif // !defined(_OpticalFlow_H_INCLUDED_)